@@ -1,5 +1,6 @@
 from typing import Union, Callable
 from functools import lru_cache
+import itertools
 import math
 from .lexer import Lexer
 from .tokens import *
@@ -13,7 +14,7 @@ class Pyrser:
     """
 
     def __init__(self) -> None:
-        self.lxr = Lexer()
+        self.lxr: Lexer = Lexer()
 
     @lru_cache(maxsize=4)
     def parse(self, re: str) -> RE:
@@ -61,13 +62,13 @@ class Pyrser:
         def parse_re() -> RE:
             return RE(parse_re_seq())
 
-        def parse_re_seq(capturing: bool = True, group_name: str = 'default') -> Union[OrNode, GroupNode]:
+        def parse_re_seq(capturing: bool = True, group_name: str = None, group_id: int = None) -> Union[OrNode, GroupNode]:
             match_start, match_end = False, False
             if type(curr_tkn) is Start or type(curr_tkn) is Circumflex:
                 next_tkn()
                 match_start = True
 
-            node = parse_group(capturing=capturing, group_name=group_name)
+            node = parse_group(capturing=capturing, group_name=group_name, group_id=group_id)
 
             if isinstance(curr_tkn, EndToken):
                 next_tkn()
@@ -82,11 +83,16 @@ class Pyrser:
 
             if isinstance(curr_tkn, OrToken):
                 next_tkn()
-                node = OrNode(left=node, right=parse_re_seq())
+                node = OrNode(left=node, right=parse_re_seq(
+                    group_name=node.group_name, group_id=node.group_id))
 
             return node
 
-        def parse_group(capturing: bool = True, group_name: str = 'default') -> GroupNode:
+        def parse_group(capturing: bool = True, group_name: str = None, group_id: int = None) -> GroupNode:
+            nonlocal groups_counter
+            if group_id is None:
+                group_id = next(groups_counter)
+
             elements = deque()  # holds the children of the GroupNode
 
             while curr_tkn is not None and not isinstance(curr_tkn, OrToken) and \
@@ -115,7 +121,7 @@ class Pyrser:
                 elements.append(new_el)
                 # next_tkn()
 
-            return GroupNode(children=elements, capturing=capturing, group_name=group_name)
+            return GroupNode(children=elements, capturing=capturing, group_name=group_name, group_id=group_id)
 
         def parse_curly(new_el: ASTNode) -> None:
             # move past the left brace
@@ -229,6 +235,7 @@ class Pyrser:
             return RangeElement(match_str="".join(sorted(set(match_str))), is_positive_logic=positive_logic)
 
         def parse_el() -> Union[Element, OrNode, GroupNode]:
+            group_name: Union[str, None]
             group_name = None
             if isinstance(curr_tkn, ElementToken):
                 return Element(match_ch=curr_tkn.char)
@@ -278,6 +285,8 @@ class Pyrser:
             next_tkn()  # consumes '>'
             return group_name
 
+        groups_counter = itertools.count(start=0)
+
         curr_tkn = None
         next_tkn = next_tkn_initializer(re)
         next_tkn()
